from __future__ import annotations

import atexit
import logging
import os
import shutil
import signal
import subprocess
import tempfile
import urllib.request
from abc import ABC, abstractmethod
from importlib.util import find_spec
from pathlib import Path
from typing import Any, Protocol

from playsound3 import backends

logger = logging.getLogger(__name__)


class PlaysoundException(Exception):
    pass


####################
## DOWNLOAD TOOLS ##
####################

_DOWNLOAD_CACHE: dict[str, str] = {}


def _download_sound_from_web(link: str, destination: Path) -> None:
    # Identifies itself as a browser to avoid HTTP 403 errors
    headers = {"User-Agent": "Mozilla/5.0 (Windows NT 6.1; Win64; x64)"}
    request = urllib.request.Request(link, headers=headers)

    with urllib.request.urlopen(request) as response, destination.open("wb") as out_file:
        out_file.write(response.read())


def _prepare_path(sound: str | Path) -> str:
    if isinstance(sound, str) and sound.startswith(("http://", "https://")):
        # To play file from URL, we download the file first to a temporary location and cache it
        if sound not in _DOWNLOAD_CACHE:
            sound_suffix = Path(sound).suffix
            with tempfile.NamedTemporaryFile(delete=False, prefix="playsound3-", suffix=sound_suffix) as f:
                _download_sound_from_web(sound, Path(f.name))
                _DOWNLOAD_CACHE[sound] = f.name
        sound = _DOWNLOAD_CACHE[sound]

    path = Path(sound)

    if not path.exists():
        raise PlaysoundException(f"file not found: {sound}")
    return path.absolute().as_posix()


########################
## BACKEND INTERFACES ##
########################


# Imitating subprocess.Popen
class PopenLike(Protocol):
    def poll(self) -> int | None: ...

    def wait(self) -> int: ...

    def terminate(self) -> None: ...


class SoundBackend(ABC):
    """Abstract class for sound backends."""

    @abstractmethod
    def check(self) -> bool:
        raise NotImplementedError("check() must be implemented.")

    @abstractmethod
    def play(self, sound: str) -> PopenLike:
        raise NotImplementedError("play() must be implemented.")


def _set_pdeathsig() -> None:
    """Set the signal delivered to this process if its parent dies."""
    try:
        import ctypes

        libc = ctypes.CDLL("libc.so.6", use_errno=True)
        PR_SET_PDEATHSIG = 1
        if libc.prctl(PR_SET_PDEATHSIG, signal.SIGTERM) != 0:
            err = ctypes.get_errno()
            raise OSError(err, os.strerror(err))
    except Exception:  # if unavailable (non-Linux) or fails, do nothing
        pass


def get_platform_specific_kwds() -> dict[str, Any]:
    """Get platform-specific keyword arguments for subprocess.Popen."""
    if os.name == "nt":
        # On Windows, this flag prevents creation of a new console window
        return {"creationflags": 0x08000000}
    else:
        # On Unix-like systems, we want to ensure that the child process is terminated if the parent process dies
        return {"preexec_fn": _set_pdeathsig}


def run_as_subprocess(commands: list[str], **kwargs: Any) -> subprocess.Popen[str]:
    """A wrapper around subprocess.Popen to handle platform-specific keyword arguments.

    By default, stdout and stderr are suppressed (set to DEVNULL).
    Additional keyword arguments can be passed and will override defaults.
    """
    popen_kwargs = {
        "stdout": subprocess.DEVNULL,
        "stderr": subprocess.DEVNULL,
        **get_platform_specific_kwds(),
        **kwargs,
    }

    return subprocess.Popen(commands, **popen_kwargs)


class Gstreamer(SoundBackend):
    """Gstreamer backend for Linux."""

    def check(self) -> bool:
        try:
            subprocess.run(
                ["gst-play-1.0", "--version"],
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
                check=True,
            )
            return True
        except FileNotFoundError:
            return False

    def play(self, sound: str) -> subprocess.Popen[str]:
        return run_as_subprocess(["gst-play-1.0", "--no-interactive", "--quiet", sound])


class Alsa(SoundBackend):
    """ALSA backend for Linux."""

    pty_master: int | None = None

    def check(self) -> bool:
        try:
            subprocess.run(["aplay", "--version"], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL, check=True)
            subprocess.run(["mpg123", "--version"], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL, check=True)
            return True
        except FileNotFoundError:
            return False

    def play(self, sound: str) -> subprocess.Popen[str]:
        suffix = Path(sound).suffix

        if self.pty_master is None:
            self.pty_master, _ = os.openpty()

        if suffix == ".wav":
            return run_as_subprocess(["aplay", "--quiet", sound])
        elif suffix == ".mp3":
            return run_as_subprocess(["mpg123", "-q", sound], stdin=self.pty_master)
        else:
            raise PlaysoundException(f"ALSA does not support for {suffix} files.")


class Ffplay(SoundBackend):
    """FFplay backend for systems with ffmpeg installed."""

    def check(self) -> bool:
        try:
            subprocess.run(["ffplay", "-version"], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL, check=True)
            return True
        except (FileNotFoundError, subprocess.CalledProcessError):
            return False

    def play(self, sound: str) -> subprocess.Popen[str]:
        return run_as_subprocess(["ffplay", "-nodisp", "-autoexit", "-loglevel", "quiet", sound])


class Wmplayer(SoundBackend):
    """Windows Media Player backend for Windows."""

    def check(self) -> bool:
        # The recommended way to check for missing library
        if find_spec("pythoncom") is None:
            return False

        try:
            import win32com.client  # type: ignore

            _ = win32com.client.Dispatch("WMPlayer.OCX")
            return True
        except (ImportError, Exception):
            # pywintypes.com_error can be raised, which inherits directly from Exception
            return False

    def play(self, sound: str) -> backends.WmplayerPopen:
        return backends.WmplayerPopen(sound)


class Winmm(SoundBackend):
    """WinMM backend for Windows."""

    def check(self) -> bool:
        try:
            import ctypes

            _ = ctypes.WinDLL("winmm.dll")  # type: ignore
            return True
        except (ImportError, FileNotFoundError, AttributeError):
            return False

    def play(self, sound: str) -> backends.WinmmPopen:
        return backends.WinmmPopen(sound)


class Afplay(SoundBackend):
    """Afplay backend for macOS."""

    def check(self) -> bool:
        # For some reason successful 'afplay -h' returns non-zero code
        # So we must use shutil to test if afplay exists
        return shutil.which("afplay") is not None

    def play(self, sound: str) -> subprocess.Popen[str]:
        return run_as_subprocess(["afplay", sound])


class Appkit(SoundBackend):
    """Appkit backend for macOS."""

    def check(self) -> bool:
        try:
            from AppKit import NSSound  # type: ignore # noqa: F401
            from Foundation import NSURL  # type: ignore # noqa: F401

            return True
        except ImportError:
            return False

    def play(self, sound: str) -> backends.AppkitPopen:
        return backends.AppkitPopen(sound)


################
## PLAYSOUND  ##
################

_NO_BACKEND_MESSAGE = "No supported audio backends on this system!"


def _auto_select_backend() -> str | None:
    if "PLAYSOUND3_BACKEND" in os.environ:
        # Allow users to override the automatic backend choice
        return os.environ["PLAYSOUND3_BACKEND"]

    for backend in _BACKEND_PREFERENCE:
        if backend in AVAILABLE_BACKENDS:
            return backend

    logger.warning(_NO_BACKEND_MESSAGE)
    return None


class Sound:
    """Subprocess-based sound object.

    Attributes:
        backend: The name of the backend used to play the sound.
        subprocess: The subprocess object used to play the sound.
    """

    def __init__(
        self,
        name: str,
        block: bool,
        backend: SoundBackend,
    ) -> None:
        """Initialize the player and begin playing."""
        self.backend: str = str(type(backend)).lower()
        self.subprocess: PopenLike = backend.play(name)

        if block:
            self.wait()

    def is_alive(self) -> bool:
        """Check if the sound is still playing.

        Returns:
            True if the sound is still playing, else False.
        """
        return self.subprocess.poll() is None

    def wait(self) -> None:
        """Block until the sound finishes playing.

        This only makes sense for non-blocking sounds.
        """
        self.subprocess.wait()

    def stop(self) -> None:
        """Stop the sound."""
        self.subprocess.terminate()


def playsound(
    sound: str | Path,
    block: bool = True,
    backend: str | None = None,
) -> Sound:
    """Play a sound file using an available audio backend.

    Args:
        sound: Path or URL of the sound file (string or pathlib.Path).
        block:
            - `True` (default): Wait until sound finishes playing.
            - `False`: Play sound in the background.
        backend: Specific audio backend to use. Leave None for automatic selection.

    Returns:
        Sound object for controlling playback.
    """
    path = _prepare_path(sound)
    backend = backend or DEFAULT_BACKEND
    if backend is None:
        raise PlaysoundException(_NO_BACKEND_MESSAGE)

    if isinstance(backend, str):
        if backend in _BACKEND_MAP:
            backend_obj = _BACKEND_MAP[backend]
        else:
            raise PlaysoundException(f"unknown backend '{backend}'")

    # Unofficially, you can pass a SoundBackend object
    elif isinstance(backend, SoundBackend):
        backend_obj = backend
    elif isinstance(backend, type) and issubclass(backend, SoundBackend):
        backend_obj = backend()
    else:
        raise PlaysoundException(f"invalid backend type '{type(backend)}'")
    return Sound(path, block, backend_obj)


def _remove_cached_downloads(cache: dict[str, str]) -> None:
    """Remove all files saved in the cache when the program ends."""
    for path in cache.values():
        Path(path).unlink()


####################
## INITIALIZATION ##
####################

atexit.register(_remove_cached_downloads, _DOWNLOAD_CACHE)

_BACKEND_PREFERENCE = [
    "gstreamer",  # Linux; should be installed on every distro
    "wmplayer",  # Windows; requires pywin32 -- should be working well on Windows
    "ffplay",  # Multiplatform; requires ffmpeg
    "appkit",  # macOS; requires PyObjC dependency
    "afplay",  # macOS; should be installed on every macOS
    "winmm",  # Windows; should be installed on every Windows, but is quirky with variable bitrate MP3s
    "alsa",  # Linux; only supports .mp3 and .wav and might not be installed
]

_BACKEND_MAP: dict[str, SoundBackend] = {
    name.lower(): obj()
    for name, obj in globals().items()
    if isinstance(obj, type) and issubclass(obj, SoundBackend) and obj is not SoundBackend
}

assert sorted(_BACKEND_PREFERENCE) == sorted(_BACKEND_MAP.keys()), "forgot to update _BACKEND_PREFERENCE?"
AVAILABLE_BACKENDS: list[str] = [name for name in _BACKEND_PREFERENCE if _BACKEND_MAP[name].check()]
DEFAULT_BACKEND: str | None = _auto_select_backend()


# This function is defined here at the bottom because of:
# SyntaxError: annotated name 'DEFAULT_BACKEND' can't be global
def prefer_backends(*backends: str) -> str | None:
    """Add backends to the top of the preference list.

    This function overrides the default backend preference.
    Backend selected here will be used ONLY if available on the system.
    This means this function can be used to update the preference for a
    specific platform without breaking the cross-platform functionality.
    After updating the preferences, the new default backend is returned.

    Args:
        backends: Names of the backends to prefer.

    Returns:
        Name of the newly selected default backend.
    """
    global DEFAULT_BACKEND, _BACKEND_PREFERENCE

    _BACKEND_PREFERENCE = list(backends) + _BACKEND_PREFERENCE
    DEFAULT_BACKEND = _auto_select_backend()
    return DEFAULT_BACKEND



import os
import time

from playsound3 import AVAILABLE_BACKENDS, playsound
from playsound3.playsound3 import _prepare_path

loc_mp3_3s = "tests/sounds/sample3s.mp3"
loc_flc_3s = "tests/sounds/sample3s.flac"
web_wav_3s = "https://samplelib.com/lib/preview/wav/sample-3s.wav"

# Download web files to the local cache
for url in [web_wav_3s]:
    _prepare_path(url)


def get_supported_sounds(backend):
    not_supporting_flac = ["alsa", "winmm"]

    if backend in not_supporting_flac:
        return [loc_mp3_3s, web_wav_3s]
    else:
        return [loc_mp3_3s, loc_flc_3s, web_wav_3s]


CI = os.environ.get("CI", False)


def test_blocking_1():
    for backend in AVAILABLE_BACKENDS:
        for path in get_supported_sounds(backend):
            t0 = time.perf_counter()
            sound = playsound(path, block=True, backend=backend)

            td = time.perf_counter() - t0
            assert not sound.is_alive(), f"backend={backend}, path={path}"
            assert td >= 3.0, f"backend={backend}, path={path}"
            assert CI or td < 5.0, f"backend={backend}, path={path}"


def test_waiting_1():
    for backend in AVAILABLE_BACKENDS:
        for path in get_supported_sounds(backend):
            t0 = time.perf_counter()
            sound = playsound(path, block=False, backend=backend)
            assert sound.is_alive(), f"backend={backend}, path={path}"

            sound.wait()
            td = time.perf_counter() - t0
            assert not sound.is_alive(), f"backend={backend}, path={path}"
            assert td >= 3.0, f"backend={backend}, path={path}"
            assert CI or td < 5.0, f"backend={backend}, path={path}"


def test_waiting_2():
    for backend in AVAILABLE_BACKENDS:
        for path in get_supported_sounds(backend):
            sound = playsound(path, block=False, backend=backend)
            assert sound.is_alive(), f"backend={backend}, path={path}"

            time.sleep(5)
            assert not sound.is_alive(), f"backend={backend}, path={path}"


def test_stopping_1():
    for backend in AVAILABLE_BACKENDS:
        for path in get_supported_sounds(backend):
            t0 = time.perf_counter()
            sound = playsound(path, block=False, backend=backend)
            assert sound.is_alive(), f"backend={backend}, path={path}"

            time.sleep(1)
            sound.stop()
            td = time.perf_counter() - t0

            time.sleep(0.05)
            assert not sound.is_alive(), f"backend={backend}, path={path}"
            assert td >= 0.95 and td < 2.0, f"backend={backend}, path={path}"

            # Stopping again should be a no-op
            sound.stop()
            assert not sound.is_alive(), f"backend={backend}, path={path}"


def test_parallel_1():
    for backend in AVAILABLE_BACKENDS:
        for path in get_supported_sounds(backend):
            t0 = time.perf_counter()
            sounds = [playsound(path, block=False, backend=backend) for _ in range(3)]
            time.sleep(0.05)
            for sound in sounds:
                assert sound.is_alive(), f"backend={backend}"
            time.sleep(1)

            sounds[1].stop()
            time.sleep(0.05)
            assert sounds[0].is_alive(), f"backend={backend}"
            assert sounds[2].is_alive(), f"backend={backend}"
            assert not sounds[1].is_alive(), f"backend={backend}"
            time.sleep(1)

            assert sounds[0].is_alive(), f"backend={backend}"
            assert sounds[2].is_alive(), f"backend={backend}"
            sounds[0].stop()
            sounds[2].stop()
            td = time.perf_counter() - t0

            time.sleep(0.05)
            for sound in sounds:
                assert not sound.is_alive(), f"backend={backend}"
            assert td >= 2.0 and td < 3.0, f"backend={backend}"


def test_parallel_2():
    N_PARALLEL = 10  # Careful - this might be loud!

    for backend in AVAILABLE_BACKENDS:
        for path in get_supported_sounds(backend):
            sounds = [playsound(path, block=False, backend=backend) for _ in range(N_PARALLEL)]

            time.sleep(1)
            for sound in sounds:
                assert sound.is_alive(), f"backend={backend}, path={path}"
            for sound in sounds:
                sound.stop()

            time.sleep(0.05)
            for sound in sounds:
                assert not sound.is_alive(), f"backend={backend}, path={path}"


def test_parallel_3():
    for backend in AVAILABLE_BACKENDS:
        sounds = [playsound(path, block=False, backend=backend) for path in get_supported_sounds(backend)]

        time.sleep(1)
        for sound in sounds:
            assert sound.is_alive(), f"backend={backend}"
        for sound in sounds:
            sound.stop()

        time.sleep(0.05)
        for sound in sounds:
            assert not sound.is_alive(), f"backend={backend}"



import urllib.error

import pytest

from playsound3 import playsound
from playsound3.playsound3 import PlaysoundException

valid = "tests/sounds/sample3s.mp3"


def test_invalid_sound_file():
    with pytest.raises(PlaysoundException):
        playsound("invalid.mp3")


def test_non_existent_file():
    with pytest.raises(PlaysoundException):
        playsound("non_existent.mp3")


def test_invalid_backend():
    with pytest.raises(PlaysoundException):
        playsound(valid, backend="invalid_backend")


def test_playsound_from_url():
    url = "https://wrong-url.com/wrong-audio.mp3"
    with pytest.raises(urllib.error.URLError):
        playsound(url)



__license__ = "MIT"
__version__ = "3.3.1"
__author__ = "Szymon Mikler"

from playsound3.playsound3 import (
    AVAILABLE_BACKENDS,
    DEFAULT_BACKEND,
    PlaysoundException,
    playsound,
    prefer_backends,
)

__all__ = [
    "AVAILABLE_BACKENDS",
    "DEFAULT_BACKEND",
    "playsound",
    "prefer_backends",
    "PlaysoundException",
]



import time

from playsound3 import AVAILABLE_BACKENDS, playsound

wav = "tests/sounds/звук 音 聲音.wav"


def test_with_blocking():
    for backend in AVAILABLE_BACKENDS:
        print(f"Testing backend: {backend}")

        sound = playsound(wav, block=True, backend=backend)
        assert not sound.is_alive()


def test_non_blocking():
    for backend in AVAILABLE_BACKENDS:
        print(f"Testing backend: {backend}")

        sound = playsound(wav, block=False, backend=backend)
        time.sleep(0.05)
        assert sound.is_alive()

        sound.wait()
        assert not sound.is_alive()



import os
import pathlib
import subprocess
import sys
import time
import typing as t

import pytest

from playsound3 import AVAILABLE_BACKENDS
from playsound3.playsound3 import _prepare_path

loc_mp3_3s = "tests/sounds/sample3s.mp3"
loc_flc_3s = "tests/sounds/sample3s.flac"
web_wav_3s = "https://samplelib.com/lib/preview/wav/sample-3s.wav"

# Download web files to the local cache
for url in [web_wav_3s]:
    _prepare_path(url)


def get_supported_sounds(backend: str):
    not_supporting_flac = ["alsa", "winmm"]
    if backend in not_supporting_flac:
        return [loc_mp3_3s, web_wav_3s]
    else:
        return [loc_mp3_3s, loc_flc_3s, web_wav_3s]


def _iter_pids() -> t.Iterable[int]:
    proc = pathlib.Path("/proc")
    for p in proc.iterdir():
        if p.name.isdigit():
            yield int(p.name)


def _read_file(path: pathlib.Path) -> bytes:
    try:
        return path.read_bytes()
    except Exception:
        return b""


def list_tagged_player_pids(tag: str) -> t.List[int]:
    """Return PIDs whose environ contains TAG=<tag>"""
    pids = []
    for pid in _iter_pids():
        base = pathlib.Path(f"/proc/{pid}")
        env = _read_file(base / "environ")
        if not env:
            continue
        # /proc/<pid>/environ is NUL-separated key=val entries
        if f"PLAYSOUND_TEST_TAG={tag}".encode() not in env.split(b"\x00"):
            continue
        cmdline = _read_file(base / "cmdline").replace(b"\x00", b" ").lower()
        if not cmdline:
            continue
        pids.append(pid)
    return pids


HELPER_CODE = """
import os, sys, time
from playsound3 import playsound

sound = playsound({path!r}, block=False, backend={backend!r})
time.sleep(10)
"""


@pytest.mark.skipif(sys.platform != "linux", reason="Linux-only: relies on /proc and PDEATHSIG semantics")
def test_killing_parent():
    TAG = "__test_killing_tag"

    for backend in AVAILABLE_BACKENDS:
        for path in get_supported_sounds(backend):
            assert len(list_tagged_player_pids(TAG)) == 0
            code = HELPER_CODE.format(path=path, backend=backend)

            environ = os.environ.copy()
            environ["PLAYSOUND_TEST_TAG"] = TAG
            proc = subprocess.Popen(["python", "-c", code], env=environ)

            time.sleep(2.5)
            assert len(list_tagged_player_pids(TAG)) == 2

            proc.kill()
            assert len(list_tagged_player_pids(TAG)) == 0
